@@ -11,7 +11,7 @@ from time import monotonic
 from queue import Queue, Empty
 from types import TracebackType
 from typing import Any, Callable, Dict, Iterator, List
-from typing import Optional, Type
+from typing import Optional, Sequence, Type
 from weakref import ref
 from contextlib import contextmanager
 
@@ -32,6 +32,7 @@ class ConnectionPool(BasePool[Connection[Any]]):
         self,
         conninfo: str = "",
         *,
+        open: bool = True,
         connection_class: Type[Connection[Any]] = Connection,
         configure: Optional[Callable[[Connection[Any]], None]] = None,
         reset: Optional[Callable[[Connection[Any]], None]] = None,
@@ -48,35 +49,14 @@ class ConnectionPool(BasePool[Connection[Any]]):
         self._pool_full_event: Optional[threading.Event] = None
 
         self._sched = Scheduler()
+        self._sched_runner: Optional[threading.Thread] = None
         self._tasks: "Queue[MaintenanceTask]" = Queue()
         self._workers: List[threading.Thread] = []
 
         super().__init__(conninfo, **kwargs)
 
-        self._sched_runner = threading.Thread(
-            target=self._sched.run, name=f"{self.name}-scheduler", daemon=True
-        )
-        for i in range(self.num_workers):
-            t = threading.Thread(
-                target=self.worker,
-                args=(self._tasks,),
-                name=f"{self.name}-worker-{i}",
-                daemon=True,
-            )
-            self._workers.append(t)
-
-        # The object state is complete. Start the worker threads
-        self._sched_runner.start()
-        for t in self._workers:
-            t.start()
-
-        # populate the pool with initial min_size connections in background
-        for i in range(self._nconns):
-            self.run_task(AddConnection(self))
-
-        # Schedule a task to shrink the pool if connections over min_size have
-        # remained unused.
-        self.schedule_task(ShrinkPool(self), self.max_idle)
+        if open:
+            self.open()
 
     def __del__(self) -> None:
         # If the '_closed' property is not set we probably failed in __init__.
@@ -84,22 +64,14 @@ class ConnectionPool(BasePool[Connection[Any]]):
         if getattr(self, "_closed", True):
             return
 
-        # Things we can try to do on a best-effort basis while the world
-        # is crumbling (a-la Eternal Sunshine of the Spotless Mind)
-        # At worse we put an item in a queue that is being deleted.
-
-        # Stop the scheduler
-        self._sched.enter(0, None)
-
-        # Stop the worker threads
-        for i in range(len(self._workers)):
-            self.run_task(StopWorker(self))
+        self._stop_workers()
 
     def wait(self, timeout: float = 30.0) -> None:
         """
         Wait for the pool to be full (with `min_size` connections) after creation.
 
-        Raise `PoolTimeout` if not ready within *timeout* sec.
+        Close the pool, and raise `PoolTimeout`, if not ready within *timeout*
+        sec.
 
         Calling this method is not mandatory: you can try and use the pool
         immediately after its creation. The first client will be served as soon
@@ -118,9 +90,7 @@ class ConnectionPool(BasePool[Connection[Any]]):
         logger.info("waiting for pool %r initialization", self.name)
         if not self._pool_full_event.wait(timeout):
             self.close()  # stop all the threads
-            raise PoolTimeout(
-                f"pool initialization incomplete after {timeout} sec"
-            )
+            raise PoolTimeout(f"pool initialization incomplete after {timeout} sec")
 
         with self._lock:
             assert self._pool_full_event
@@ -129,9 +99,7 @@ class ConnectionPool(BasePool[Connection[Any]]):
         logger.info("pool %r is ready to use", self.name)
 
     @contextmanager
-    def connection(
-        self, timeout: Optional[float] = None
-    ) -> Iterator[Connection[Any]]:
+    def connection(self, timeout: Optional[float] = None) -> Iterator[Connection[Any]]:
         """Context manager to obtain a connection from the pool.
 
         Return the connection immediately if available, otherwise wait up to
@@ -142,7 +110,6 @@ class ConnectionPool(BasePool[Connection[Any]]):
         :ref:`connection context behaviour <with-connection>` (commit/rollback
         the transaction in case of success/error). If the connection is no more
         in working state replace it with a new one.
-
         """
         conn = self.getconn(timeout=timeout)
         t0 = monotonic()
@@ -155,9 +122,9 @@ class ConnectionPool(BasePool[Connection[Any]]):
             self.putconn(conn)
 
     def getconn(self, timeout: Optional[float] = None) -> Connection[Any]:
-        """Obtain a contection from the pool.
+        """Obtain a connection from the pool.
 
-        You should preferrably use `connection()`. Use this function only if
+        You should preferably use `connection()`. Use this function only if
         it is not possible to use the connection as context manager.
 
         After using this function you *must* call a corresponding `putconn()`:
@@ -171,21 +138,8 @@ class ConnectionPool(BasePool[Connection[Any]]):
         # or if the client needs to wait.
         with self._lock:
             self._check_open_getconn()
-
-            pos: Optional[WaitingClient] = None
-            if self._pool:
-                # Take a connection ready out of the pool
-                conn = self._pool.popleft()
-                if len(self._pool) < self._nconns_min:
-                    self._nconns_min = len(self._pool)
-            else:
-                if self.max_waiting and len(self._waiting) >= self.max_waiting:
-                    self._stats[self._REQUESTS_ERRORS] += 1
-                    raise TooManyRequests(
-                        f"the pool {self.name!r} has aleady"
-                        f" {len(self._waiting)} requests waiting"
-                    )
-
+            conn = self._get_ready_connection(timeout)
+            if not conn:
                 # No connection available: put the client in the waiting queue
                 t0 = monotonic()
                 pos = WaitingClient()
@@ -193,19 +147,11 @@ class ConnectionPool(BasePool[Connection[Any]]):
                 self._stats[self._REQUESTS_QUEUED] += 1
 
                 # If there is space for the pool to grow, let's do it
-                # Allow only one thread at time to grow the pool (or returning
-                # connections might be starved).
-                if self._nconns < self._max_size and not self._growing:
-                    self._nconns += 1
-                    logger.info(
-                        "growing pool %r to %s", self.name, self._nconns
-                    )
-                    self._growing = True
-                    self.run_task(AddConnection(self, growing=True))
+                self._maybe_grow_pool()
 
         # If we are in the waiting queue, wait to be assigned a connection
         # (outside the critical section, so only the waiting client is locked)
-        if pos:
+        if not conn:
             if timeout is None:
                 timeout = self.timeout
             try:
@@ -224,6 +170,34 @@ class ConnectionPool(BasePool[Connection[Any]]):
         logger.info("connection given by %r", self.name)
         return conn
 
+    def _get_ready_connection(
+        self, timeout: Optional[float]
+    ) -> Optional[Connection[Any]]:
+        """Return a connection, if the client deserves one."""
+        conn: Optional[Connection[Any]] = None
+        if self._pool:
+            # Take a connection ready out of the pool
+            conn = self._pool.popleft()
+            if len(self._pool) < self._nconns_min:
+                self._nconns_min = len(self._pool)
+        elif self.max_waiting and len(self._waiting) >= self.max_waiting:
+            self._stats[self._REQUESTS_ERRORS] += 1
+            raise TooManyRequests(
+                f"the pool {self.name!r} has already"
+                f" {len(self._waiting)} requests waiting"
+            )
+        return conn
+
+    def _maybe_grow_pool(self) -> None:
+        # Allow only one thread at time to grow the pool (or returning
+        # connections might be starved).
+        if self._nconns >= self._max_size or self._growing:
+            return
+        self._nconns += 1
+        logger.info("growing pool %r to %s", self.name, self._nconns)
+        self._growing = True
+        self.run_task(AddConnection(self, growing=True))
+
     def putconn(self, conn: Connection[Any]) -> None:
         """Return a connection to the loving hands of its pool.
 
@@ -235,11 +209,7 @@ class ConnectionPool(BasePool[Connection[Any]]):
 
         logger.info("returning connection to %r", self.name)
 
-        # If the pool is closed just close the connection instead of returning
-        # it to the pool. For extra refcare remove the pool reference from it.
-        if self._closed:
-            conn._pool = None
-            conn.close()
+        if self._maybe_close_connection(conn):
             return
 
         # Use a worker to perform eventual maintenance work in a separate thread
@@ -248,10 +218,85 @@ class ConnectionPool(BasePool[Connection[Any]]):
         else:
             self._return_connection(conn)
 
+    def _maybe_close_connection(self, conn: Connection[Any]) -> bool:
+        """Close a returned connection if necessary.
+
+        Return `!True if the connection was closed.
+        """
+        # If the pool is closed just close the connection instead of returning
+        # it to the pool. For extra refcare remove the pool reference from it.
+        if not self._closed:
+            return False
+
+        conn._pool = None
+        conn.close()
+        return True
+
+    def open(self, wait: bool = False, timeout: float = 30.0) -> None:
+        """Open the pool by starting connecting and and accepting clients.
+
+        If *wait* is `!False`, return immediately and let the background worker
+        fill the pool if `min_size` > 0. Otherwise wait up to *timeout* seconds
+        for the requested number of connections to be ready (see `wait()` for
+        details).
+
+        It is safe to call `!open()` again on a pool already open (because the
+        method was already called, or because the pool context was entered, or
+        because the pool was initialized with *open* = `!True`) but you cannot
+        currently re-open a closed pool.
+        """
+        with self._lock:
+            self._open()
+
+        if wait:
+            self.wait(timeout=timeout)
+
+    def _open(self) -> None:
+        if not self._closed:
+            return
+
+        self._check_open()
+
+        self._start_workers()
+        self._start_initial_tasks()
+
+        self._closed = False
+        self._opened = True
+
+    def _start_workers(self) -> None:
+        self._sched_runner = threading.Thread(
+            target=self._sched.run,
+            name=f"{self.name}-scheduler",
+            daemon=True,
+        )
+        assert not self._workers
+        for i in range(self.num_workers):
+            t = threading.Thread(
+                target=self.worker,
+                args=(self._tasks,),
+                name=f"{self.name}-worker-{i}",
+                daemon=True,
+            )
+            self._workers.append(t)
+
+        # The object state is complete. Start the worker threads
+        self._sched_runner.start()
+        for t in self._workers:
+            t.start()
+
+    def _start_initial_tasks(self) -> None:
+        # populate the pool with initial min_size connections in background
+        for i in range(self._nconns):
+            self.run_task(AddConnection(self))
+
+        # Schedule a task to shrink the pool if connections over min_size have
+        # remained unused.
+        self.schedule_task(ShrinkPool(self), self.max_idle)
+
     def close(self, timeout: float = 5.0) -> None:
         """Close the pool and make it unavailable to new clients.
 
-        All the waiting and future client will fail to acquire a connection
+        All the waiting and future clients will fail to acquire a connection
         with a `PoolClosed` exception. Currently used connections will not be
         closed until returned to the pool.
 
@@ -269,30 +314,41 @@ class ConnectionPool(BasePool[Connection[Any]]):
             # Take waiting client and pool connections out of the state
             waiting = list(self._waiting)
             self._waiting.clear()
-            pool = list(self._pool)
+            connections = list(self._pool)
             self._pool.clear()
 
         # Now that the flag _closed is set, getconn will fail immediately,
         # putconn will just close the returned connection.
+        self._stop_workers(waiting, connections, timeout)
+
+    def _stop_workers(
+        self,
+        waiting_clients: Sequence["WaitingClient"] = (),
+        connections: Sequence[Connection[Any]] = (),
+        timeout: float = 0.0,
+    ) -> None:
 
         # Stop the scheduler
         self._sched.enter(0, None)
 
         # Stop the worker threads
-        for i in range(len(self._workers)):
+        workers, self._workers = self._workers[:], []
+        for i in range(len(workers)):
             self.run_task(StopWorker(self))
 
         # Signal to eventual clients in the queue that business is closed.
-        for pos in waiting:
+        for pos in waiting_clients:
             pos.fail(PoolClosed(f"the pool {self.name!r} is closed"))
 
         # Close the connections still in the pool
-        for conn in pool:
+        for conn in connections:
             conn.close()
 
         # Wait for the worker threads to terminate
+        assert self._sched_runner is not None
+        sched_runner, self._sched_runner = self._sched_runner, None
         if timeout > 0:
-            for t in [self._sched_runner] + self._workers:
+            for t in [sched_runner] + workers:
                 if not t.is_alive():
                     continue
                 t.join(timeout)
@@ -305,6 +361,7 @@ class ConnectionPool(BasePool[Connection[Any]]):
                     )
 
     def __enter__(self) -> "ConnectionPool":
+        self.open()
         return self
 
     def __exit__(
@@ -386,7 +443,7 @@ class ConnectionPool(BasePool[Connection[Any]]):
         # Don't make all the workers time out at the same moment
         timeout = cls._jitter(cls._WORKER_TIMEOUT, -0.1, 0.1)
         while True:
-            # Use a timeout to make the wait interruptable
+            # Use a timeout to make the wait interruptible
             try:
                 task = q.get(timeout=timeout)
             except Empty:
@@ -410,13 +467,17 @@ class ConnectionPool(BasePool[Connection[Any]]):
                     ex,
                 )
 
-    def _connect(self) -> Connection[Any]:
+    def _connect(self, timeout: Optional[float] = None) -> Connection[Any]:
         """Return a new connection configured for the pool."""
         self._stats[self._CONNECTIONS_NUM] += 1
+        kwargs = self.kwargs
+        if timeout:
+            kwargs = kwargs.copy()
+            kwargs["connect_timeout"] = max(round(timeout), 1)
         t0 = monotonic()
         try:
             conn: Connection[Any]
-            conn = self.connection_class.connect(self.conninfo, **self.kwargs)
+            conn = self.connection_class.connect(self.conninfo, **kwargs)
         except Exception:
             self._stats[self._CONNECTIONS_ERRORS] += 1
             raise
@@ -452,9 +513,7 @@ class ConnectionPool(BasePool[Connection[Any]]):
         """
         now = monotonic()
         if not attempt:
-            attempt = ConnectionAttempt(
-                reconnect_timeout=self.reconnect_timeout
-            )
+            attempt = ConnectionAttempt(reconnect_timeout=self.reconnect_timeout)
 
         try:
             conn = self._connect()
@@ -483,9 +542,7 @@ class ConnectionPool(BasePool[Connection[Any]]):
             with self._lock:
                 if self._nconns < self._max_size and self._waiting:
                     self._nconns += 1
-                    logger.info(
-                        "growing pool %r to %s", self.name, self._nconns
-                    )
+                    logger.info("growing pool %r to %s", self.name, self._nconns)
                     self.run_task(AddConnection(self, growing=True))
                 else:
                     self._growing = False
@@ -683,9 +740,7 @@ class MaintenanceTask(ABC):
 
     def __init__(self, pool: "ConnectionPool"):
         self.pool = ref(pool)
-        logger.debug(
-            "task created in %s: %s", threading.current_thread().name, self
-        )
+        logger.debug("task created in %s: %s", threading.current_thread().name, self)
 
     def __repr__(self) -> str:
         pool = self.pool()
@@ -703,9 +758,7 @@ class MaintenanceTask(ABC):
             # Pool is no more working. Quietly discard the operation.
             return
 
-        logger.debug(
-            "task running in %s: %s", threading.current_thread().name, self
-        )
+        logger.debug("task running in %s: %s", threading.current_thread().name, self)
         self._run(pool)
 
     def tick(self) -> None:
