@@ -4,24 +4,24 @@ psycopg asynchronous connection pool
 
 # Copyright (C) 2021 The Psycopg Team
 
-import sys
 import asyncio
 import logging
 from abc import ABC, abstractmethod
 from time import monotonic
 from types import TracebackType
 from typing import Any, AsyncIterator, Awaitable, Callable
-from typing import Dict, List, Optional, Type
+from typing import Dict, List, Optional, Sequence, Type
 from weakref import ref
+from contextlib import asynccontextmanager
 
 from psycopg import errors as e
+from psycopg import AsyncConnection
 from psycopg.pq import TransactionStatus
-from psycopg.connection_async import AsyncConnection
 
 from .base import ConnectionAttempt, BasePool
 from .sched import AsyncScheduler
 from .errors import PoolClosed, PoolTimeout, TooManyRequests
-from ._compat import Task, asynccontextmanager, create_task, Deque
+from ._compat import Task, create_task, Deque
 
 logger = logging.getLogger("psycopg.pool")
 
@@ -31,21 +31,12 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         self,
         conninfo: str = "",
         *,
+        open: bool = True,
         connection_class: Type[AsyncConnection[Any]] = AsyncConnection,
-        configure: Optional[
-            Callable[[AsyncConnection[Any]], Awaitable[None]]
-        ] = None,
-        reset: Optional[
-            Callable[[AsyncConnection[Any]], Awaitable[None]]
-        ] = None,
+        configure: Optional[Callable[[AsyncConnection[Any]], Awaitable[None]]] = None,
+        reset: Optional[Callable[[AsyncConnection[Any]], Awaitable[None]]] = None,
         **kwargs: Any,
     ):
-        # https://bugs.python.org/issue42600
-        if sys.version_info < (3, 7):
-            raise e.NotSupportedError(
-                "async pool not supported before Python 3.7"
-            )
-
         self.connection_class = connection_class
         self._configure = configure
         self._reset = reset
@@ -57,28 +48,14 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         self._pool_full_event: Optional[asyncio.Event] = None
 
         self._sched = AsyncScheduler()
+        self._sched_runner: Optional[Task[None]] = None
         self._tasks: "asyncio.Queue[MaintenanceTask]" = asyncio.Queue()
         self._workers: List[Task[None]] = []
 
         super().__init__(conninfo, **kwargs)
 
-        self._sched_runner = create_task(
-            self._sched.run(), name=f"{self.name}-scheduler"
-        )
-        for i in range(self.num_workers):
-            t = create_task(
-                self.worker(self._tasks),
-                name=f"{self.name}-worker-{i}",
-            )
-            self._workers.append(t)
-
-        # populate the pool with initial min_size connections in background
-        for i in range(self._nconns):
-            self.run_task(AddConnection(self))
-
-        # Schedule a task to shrink the pool if connections over min_size have
-        # remained unused.
-        self.run_task(Schedule(self, ShrinkPool(self), self.max_idle))
+        if open:
+            self._open()
 
     async def wait(self, timeout: float = 30.0) -> None:
         self._check_open_getconn()
@@ -93,7 +70,7 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         try:
             await asyncio.wait_for(self._pool_full_event.wait(), timeout)
         except asyncio.TimeoutError:
-            await self.close()  # stop all the threads
+            await self.close()  # stop all the tasks
             raise PoolTimeout(
                 f"pool initialization incomplete after {timeout} sec"
             ) from None
@@ -118,9 +95,7 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
             self._stats[self._USAGE_MS] += int(1000.0 * (t1 - t0))
             await self.putconn(conn)
 
-    async def getconn(
-        self, timeout: Optional[float] = None
-    ) -> AsyncConnection[Any]:
+    async def getconn(self, timeout: Optional[float] = None) -> AsyncConnection[Any]:
         logger.info("connection requested from %r", self.name)
         self._stats[self._REQUESTS_NUM] += 1
 
@@ -128,40 +103,20 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         # or if the client needs to wait.
         async with self._lock:
             self._check_open_getconn()
-
-            pos: Optional[AsyncClient] = None
-            if self._pool:
-                # Take a connection ready out of the pool
-                conn = self._pool.popleft()
-                if len(self._pool) < self._nconns_min:
-                    self._nconns_min = len(self._pool)
-            else:
-                if self.max_waiting and len(self._waiting) >= self.max_waiting:
-                    self._stats[self._REQUESTS_ERRORS] += 1
-                    raise TooManyRequests(
-                        f"the pool {self.name!r} has aleady"
-                        f" {len(self._waiting)} requests waiting"
-                    )
-
+            conn = await self._get_ready_connection(timeout)
+            if not conn:
                 # No connection available: put the client in the waiting queue
                 t0 = monotonic()
                 pos = AsyncClient()
                 self._waiting.append(pos)
                 self._stats[self._REQUESTS_QUEUED] += 1
 
-                # Allow only one thread at time to grow the pool (or returning
-                # connections might be starved).
-                if self._nconns < self._max_size and not self._growing:
-                    self._nconns += 1
-                    logger.info(
-                        "growing pool %r to %s", self.name, self._nconns
-                    )
-                    self._growing = True
-                    self.run_task(AddConnection(self, growing=True))
+                # If there is space for the pool to grow, let's do it
+                self._maybe_grow_pool()
 
         # If we are in the waiting queue, wait to be assigned a connection
         # (outside the critical section, so only the waiting client is locked)
-        if pos:
+        if not conn:
             if timeout is None:
                 timeout = self.timeout
             try:
@@ -180,24 +135,94 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         logger.info("connection given by %r", self.name)
         return conn
 
+    async def _get_ready_connection(
+        self, timeout: Optional[float]
+    ) -> Optional[AsyncConnection[Any]]:
+        conn: Optional[AsyncConnection[Any]] = None
+        if self._pool:
+            # Take a connection ready out of the pool
+            conn = self._pool.popleft()
+            if len(self._pool) < self._nconns_min:
+                self._nconns_min = len(self._pool)
+        elif self.max_waiting and len(self._waiting) >= self.max_waiting:
+            self._stats[self._REQUESTS_ERRORS] += 1
+            raise TooManyRequests(
+                f"the pool {self.name!r} has already"
+                f" {len(self._waiting)} requests waiting"
+            )
+        return conn
+
+    def _maybe_grow_pool(self) -> None:
+        # Allow only one task at time to grow the pool (or returning
+        # connections might be starved).
+        if self._nconns < self._max_size and not self._growing:
+            self._nconns += 1
+            logger.info("growing pool %r to %s", self.name, self._nconns)
+            self._growing = True
+            self.run_task(AddConnection(self, growing=True))
+
     async def putconn(self, conn: AsyncConnection[Any]) -> None:
         self._check_pool_putconn(conn)
 
         logger.info("returning connection to %r", self.name)
-
-        # If the pool is closed just close the connection instead of returning
-        # it to the pool. For extra refcare remove the pool reference from it.
-        if self._closed:
-            conn._pool = None
-            await conn.close()
+        if await self._maybe_close_connection(conn):
             return
 
-        # Use a worker to perform eventual maintenance work in a separate thread
+        # Use a worker to perform eventual maintenance work in a separate task
         if self._reset:
             self.run_task(ReturnConnection(self, conn))
         else:
             await self._return_connection(conn)
 
+    async def _maybe_close_connection(self, conn: AsyncConnection[Any]) -> bool:
+        # If the pool is closed just close the connection instead of returning
+        # it to the pool. For extra refcare remove the pool reference from it.
+        if not self._closed:
+            return False
+
+        conn._pool = None
+        await conn.close()
+        return True
+
+    async def open(self, wait: bool = False, timeout: float = 30.0) -> None:
+        async with self._lock:
+            self._open()
+
+        if wait:
+            await self.wait(timeout=timeout)
+
+    def _open(self) -> None:
+        if not self._closed:
+            return
+
+        self._check_open()
+
+        self._start_workers()
+        self._start_initial_tasks()
+
+        self._closed = False
+        self._opened = True
+
+    def _start_workers(self) -> None:
+        self._sched_runner = create_task(
+            self._sched.run(), name=f"{self.name}-scheduler"
+        )
+        for i in range(self.num_workers):
+            t = create_task(
+                self.worker(self._tasks),
+                name=f"{self.name}-worker-{i}",
+            )
+            self._workers.append(t)
+
+    def _start_initial_tasks(self) -> None:
+        # populate the pool with initial min_size connections in background
+        for i in range(self._nconns):
+            self.run_task(AddConnection(self))
+
+        # Schedule a task to shrink the pool if connections over min_size have
+        # remained unused.
+        self.run_task(Schedule(self, ShrinkPool(self), self.max_idle))
+
     async def close(self, timeout: float = 5.0) -> None:
         if self._closed:
             return
@@ -209,29 +234,39 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
             # Take waiting client and pool connections out of the state
             waiting = list(self._waiting)
             self._waiting.clear()
-            pool = list(self._pool)
+            connections = list(self._pool)
             self._pool.clear()
 
         # Now that the flag _closed is set, getconn will fail immediately,
         # putconn will just close the returned connection.
+        await self._stop_workers(waiting, connections, timeout)
 
+    async def _stop_workers(
+        self,
+        waiting_clients: Sequence["AsyncClient"] = (),
+        connections: Sequence[AsyncConnection[Any]] = (),
+        timeout: float = 0.0,
+    ) -> None:
         # Stop the scheduler
         await self._sched.enter(0, None)
 
-        # Stop the worker threads
-        for w in self._workers:
+        # Stop the worker tasks
+        workers, self._workers = self._workers[:], []
+        for w in workers:
             self.run_task(StopWorker(self))
 
         # Signal to eventual clients in the queue that business is closed.
-        for pos in waiting:
+        for pos in waiting_clients:
             await pos.fail(PoolClosed(f"the pool {self.name!r} is closed"))
 
         # Close the connections still in the pool
-        for conn in pool:
+        for conn in connections:
             await conn.close()
 
-        # Wait for the worker threads to terminate
-        wait = asyncio.gather(self._sched_runner, *self._workers)
+        # Wait for the worker tasks to terminate
+        assert self._sched_runner is not None
+        sched_runner, self._sched_runner = self._sched_runner, None
+        wait = asyncio.gather(sched_runner, *workers)
         try:
             if timeout > 0:
                 await asyncio.wait_for(asyncio.shield(wait), timeout=timeout)
@@ -245,6 +280,7 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
             )
 
     async def __aenter__(self) -> "AsyncConnectionPool":
+        await self.open()
         return self
 
     async def __aexit__(
@@ -255,9 +291,7 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
     ) -> None:
         await self.close()
 
-    async def resize(
-        self, min_size: int, max_size: Optional[int] = None
-    ) -> None:
+    async def resize(self, min_size: int, max_size: Optional[int] = None) -> None:
         min_size, max_size = self._check_size(min_size, max_size)
 
         ngrow = max(0, min_size - self._min_size)
@@ -301,20 +335,18 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         self._reconnect_failed(self)
 
     def run_task(self, task: "MaintenanceTask") -> None:
-        """Run a maintenance task in a worker thread."""
+        """Run a maintenance task in a worker."""
         self._tasks.put_nowait(task)
 
-    async def schedule_task(
-        self, task: "MaintenanceTask", delay: float
-    ) -> None:
-        """Run a maintenance task in a worker thread in the future."""
+    async def schedule_task(self, task: "MaintenanceTask", delay: float) -> None:
+        """Run a maintenance task in a worker in the future."""
         await self._sched.enter(delay, task.tick)
 
     @classmethod
     async def worker(cls, q: "asyncio.Queue[MaintenanceTask]") -> None:
         """Runner to execute pending maintenance task.
 
-        The function is designed to run as a separate thread.
+        The function is designed to run as a task.
 
         Block on the queue *q*, run a task received. Finish running if a
         StopWorker is received.
@@ -337,15 +369,16 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
                     ex,
                 )
 
-    async def _connect(self) -> AsyncConnection[Any]:
-        """Return a new connection configured for the pool."""
+    async def _connect(self, timeout: Optional[float] = None) -> AsyncConnection[Any]:
         self._stats[self._CONNECTIONS_NUM] += 1
+        kwargs = self.kwargs
+        if timeout:
+            kwargs = kwargs.copy()
+            kwargs["connect_timeout"] = max(round(timeout), 1)
         t0 = monotonic()
         try:
             conn: AsyncConnection[Any]
-            conn = await self.connection_class.connect(
-                self.conninfo, **self.kwargs
-            )
+            conn = await self.connection_class.connect(self.conninfo, **kwargs)
         except Exception:
             self._stats[self._CONNECTIONS_ERRORS] += 1
             raise
@@ -381,9 +414,7 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         """
         now = monotonic()
         if not attempt:
-            attempt = ConnectionAttempt(
-                reconnect_timeout=self.reconnect_timeout
-            )
+            attempt = ConnectionAttempt(reconnect_timeout=self.reconnect_timeout)
 
         try:
             conn = await self._connect()
@@ -412,9 +443,7 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
             async with self._lock:
                 if self._nconns < self._max_size and self._waiting:
                     self._nconns += 1
-                    logger.info(
-                        "growing pool %r to %s", self.name, self._nconns
-                    )
+                    logger.info("growing pool %r to %s", self.name, self._nconns)
                     self.run_task(AddConnection(self, growing=True))
                 else:
                     self._growing = False
@@ -623,7 +652,7 @@ class MaintenanceTask(ABC):
     async def run(self) -> None:
         """Run the task.
 
-        This usually happens in a worker thread. Call the concrete _run()
+        This usually happens in a worker. Call the concrete _run()
         implementation, if the pool is still alive.
         """
         pool = self.pool()
@@ -636,7 +665,7 @@ class MaintenanceTask(ABC):
     async def tick(self) -> None:
         """Run the scheduled task
 
-        This function is called by the scheduler thread. Use a worker to
+        This function is called by the scheduler task. Use a worker to
         run the task for real in order to free the scheduler immediately.
         """
         pool = self.pool()
@@ -652,7 +681,7 @@ class MaintenanceTask(ABC):
 
 
 class StopWorker(MaintenanceTask):
-    """Signal the maintenance thread to terminate."""
+    """Signal the maintenance worker to terminate."""
 
     async def _run(self, pool: "AsyncConnectionPool") -> None:
         pass
@@ -676,9 +705,7 @@ class AddConnection(MaintenanceTask):
 class ReturnConnection(MaintenanceTask):
     """Clean up and return a connection to the pool."""
 
-    def __init__(
-        self, pool: "AsyncConnectionPool", conn: "AsyncConnection[Any]"
-    ):
+    def __init__(self, pool: "AsyncConnectionPool", conn: "AsyncConnection[Any]"):
         super().__init__(pool)
         self.conn = conn
 
